"""
Quoridor Online
Quentin Deschamps, 2020
"""


class Game:
    """Create a game"""
    def __init__(self, game_id, nb_players):
        self.game_id = game_id
        self.nb_players = nb_players
        self.connected = [False] * nb_players
        self.names = [''] * nb_players
        self.run = False
        self.current_player = -1
        self.last_play = ''
        self.winner = ''
        self.wanted_restart = []

    def add_player(self, num_player):
        """Add a player"""
        self.connected[num_player] = True
        print(f"[Game {self.game_id}]: player {num_player} added")

    def add_name(self, data):
        """Add a player's name"""
        num_player = int(data.split(';')[1])
        name = data.split(';')[2]
        self.names[num_player] = name
        print(f"[Game {self.game_id}]: {name} added as player {num_player}")

    def remove_player(self, num_player):
        """Remove a player"""
        self.connected[num_player] = False
        self.names[num_player] = ''
        if self.nb_players_connected() == 1:
            self.run = False
            self.current_player == -1
        else:
            if num_player == self.current_player:
                self.current_player = self.next_player(self.current_player)
                self.last_play = ';'.join(['D', str(num_player)])
        print(f"[Game {self.game_id}]: player {num_player} removed")

    def nb_players_connected(self):
        """Return the number of players connected"""
        return self.connected.count(True)

    def players_missing(self):
        """Return the number of players missing to start"""
        return self.nb_players - self.nb_players_connected()

    def ready(self):
        """Return True if the game can start"""
        return self.nb_players_connected() == self.nb_players

    def next_player(self, current):
        """Return the new current player"""
        current = (current + 1) % self.nb_players
        while not self.connected[current]:
            current = (current + 1) % self.nb_players
        return current

    def get_name_current(self):
        """Return the name of the current player"""
        return self.names[self.current_player]

    def start(self):
        """Start the game"""
        self.winner = ''
        self.current_player = self.next_player(-1)
        self.run = True
        print(f"[Game {self.game_id}]: started")

    def play(self, data):
        """Get a move"""
        print(f"[Game {self.game_id}]: move {data}")
        self.last_play = data
        if data.split(';')[-1] == 'w':
            print(f"[Game {self.game_id}]: {self.get_name_current()} wins!")
            self.winner = self.get_name_current()
            self.current_player = -1
            self.run = False
            self.wanted_restart = []
        else:
            self.current_player = self.next_player(self.current_player)

    def restart(self, data):
        """Restart the game if there are enough players"""
        num_player = int(data.split(';')[1])
        self.wanted_restart.append(num_player)
        print(f"[Game {self.game_id}]: {self.names[num_player]} asked restart",
              end=' ')
        print(f"{len(self.wanted_restart)}/{self.nb_players}")
        if len(self.wanted_restart) == self.nb_players:
            self.start()


class Games:
    """Manage games"""
    def __init__(self, nb_players):
        self.games = {}
        self.nb_players = nb_players
        self.num_player = 0
        self.game_id = 0

    def find_game(self, game_id):
        """Find a game"""
        if game_id in self.games:
            return self.games[game_id]
        return None

    def add_game(self):
        """Create a new game"""
        if self.game_id not in self.games:
            self.num_player = 0
            self.games[self.game_id] = Game(self.game_id, self.nb_players)
            print(f"[Game {self.game_id}]: created")

    def del_game(self, game_id):
        """Delete a game"""
        if game_id in self.games:
            del self.games[game_id]
            print(f"[Game {game_id}]: closed")
            if game_id == self.game_id:
                self.num_player = 0

    def accept_player(self):
        """Accept a player"""
        if self.game_id not in self.games:
            self.add_game()
        self.games[self.game_id].add_player(self.num_player)
        return self.game_id, self.num_player

    def launch_game(self):
        """Lauch a game"""
        if self.games[self.game_id].ready():
            self.games[self.game_id].start()
            self.game_id += 1
        else:
            self.num_player += 1

    def remove_player(self, game_id, num_player):
        """Remove a player"""
        game = self.find_game(game_id)
        if game is not None:
            game.remove_player(num_player)
            if not game.run:
                self.del_game(game_id)



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import sys
import socket
from _thread import start_new_thread
import pickle
from quoridor.server.src.game import Games


def threaded_client(conn, nb_players, num_player, game_id, games):
    """Manage threads for clients"""
    first_data = ";".join([str(nb_players), str(num_player)])
    conn.send(str.encode(first_data))
    run = True
    while run:
        try:
            data = conn.recv(2048).decode()
            game = games.find_game(game_id)
            if game is not None:
                if not data:
                    break
                elif data.split(';')[0] == 'N':
                    game.add_name(data)
                elif data.split(';')[0] == 'P':
                    game.play(data)
                elif data.split(';')[0] == 'R':
                    game.restart(data)
                conn.sendall(pickle.dumps(game))
            else:
                run = False
        except socket.error as e:
            print(e)

    games.remove_player(game_id, num_player)
    conn.close()


def server(host, port, nb_players):
    """Game server"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        s.bind((host, port))
    except socket.error as e:
        str(e)

    s.listen(nb_players)
    print(f"Server for {nb_players} players games started!")
    print("Waiting for a connection...")

    games = Games(nb_players)

    while True:
        conn, addr = s.accept()
        print("Connected to:", addr)
        game_id, num_player = games.accept_player()
        start_new_thread(threaded_client,
                         (conn, nb_players, num_player, game_id, games))
        games.launch_game()


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Not enough arguments")
        print("[Host address] [Port number] [Number of players]")
    else:
        host = str(sys.argv[1])
        port = int(sys.argv[2])
        nb_players = int(sys.argv[3])
        if nb_players not in [2, 3, 4]:
            print("Only for 2, 3 or 4 players")
        else:
            server(host, port, nb_players)

# Get IP address on Linux: hostname -I



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame
from quoridor.client.src.colors import Colors
from quoridor.client.src.widgets import Text, Button
from quoridor.client.src.coord import Coords


def pos_in_rect(rect, pos):
    """Return True if pos is in the rectangle"""
    pos_x, pos_y = pos
    x, y, width, height = rect
    return (x <= pos_x <= x + width
            and y <= pos_y <= y + height)


class Window:
    """Create the window"""
    def __init__(self, width=1000, height=830, case_side=65, wall_width=15,
                 title="Quoridor Online", bgcolor=Colors.white):
        self.width = width
        self.height = height
        self.case_side = case_side
        self.wall_width = wall_width
        self.bgcolor = bgcolor
        self.top_left = (20, 20)
        self.side_board = 9*self.case_side + 10*self.wall_width
        self.win = pygame.display.set_mode((width, height))
        pygame.display.set_caption(title)
        self.button_restart = Button("Restart", self.side_board + 60,
                                     self.side_board - 100, Colors.red,
                                     show=False)
        self.button_quit = Button("Quit", self.side_board + 60,
                                  self.side_board - 50, Colors.red)
        self.buttons = [self.button_restart, self.button_quit]
        self.title = Text("Quoridor", Colors.black, size=50)
        self.info = Text("Welcome to Quoridor Online!", Colors.black, size=45)
        self.coords = Coords(self)

    def update_info(self, text, color=None):
        """Update info text"""
        self.info.text = text
        if color is not None:
            self.info.color = color

    def draw_game_board(self, pos):
        """Draw the game board"""
        for c in self.coords.coords:
            rect = c.rect
            if pos_in_rect(rect, pos):
                color = Colors.grey_dark
            else:
                color = Colors.grey
                wall_east = c.wall_east
                wall_south = c.wall_south
                if wall_east and pos_in_rect(wall_east.rect_small, pos):
                    wall_east.draw(Colors.grey_dark)
                elif wall_south and pos_in_rect(wall_south.rect_small, pos):
                    wall_south.draw(Colors.grey_dark)
            c.draw(color)

    def draw_finish_lines(self, players):
        """Draw the finish lines with the player's color"""
        for p in players.players:
            if p.name != '':
                if p.orient == "north":
                    pygame.draw.line(
                        self.win, p.color,
                        (self.top_left[0], self.top_left[1] + self.side_board),
                        (self.top_left[0] + self.side_board,
                         self.top_left[1] + self.side_board),
                        self.wall_width)
                elif p.orient == "east":
                    pygame.draw.line(
                        self.win, p.color,
                        (self.top_left[0], self.top_left[1]),
                        (self.top_left[0], self.top_left[1] + self.side_board),
                        self.wall_width)
                elif p.orient == "south":
                    pygame.draw.line(
                        self.win, p.color,
                        (self.top_left[0], self.top_left[1]),
                        (self.top_left[0] + self.side_board, self.top_left[1]),
                        self.wall_width)
                elif p.orient == "west":
                    pygame.draw.line(
                        self.win, p.color,
                        (self.top_left[0] + self.side_board, self.top_left[1]),
                        (self.top_left[0] + self.side_board,
                         self.top_left[1] + self.side_board),
                        self.wall_width)

    def draw_right_panel(self, game, players):
        """Draw the right panel with player's informations"""
        x, y = self.side_board + 50, 20
        self.title.draw(self.win, (x + 10, y))
        for p in players.players:
            if p.name != '':
                text_p = Text(f"{p.name}: {p.walls_remain} walls", p.color)
                text_p.draw(self.win, (x, y + 100*p.num_player + 100))

    def draw_buttons(self):
        """Draw buttons"""
        for b in self.buttons:
            if b.show:
                b.draw(self.win)

    def redraw_window(self, game, players, walls, pos):
        """Redraw the full window"""
        self.win.fill(self.bgcolor)
        self.draw_game_board(pos)
        self.draw_finish_lines(players)
        self.draw_right_panel(game, players)
        self.draw_buttons()
        players.draw(self)
        walls.draw()
        self.info.draw(self.win, (self.top_left[0], self.height - 50))
        pygame.display.update()



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame


class Sounds:
    """Manage sounds"""
    def __init__(self, path):
        pygame.mixer.init()
        self.start_sound = pygame.mixer.Sound(
            "".join([path, "/sounds/start_sound.wav"]))
        self.winning_sound = pygame.mixer.Sound(
            "".join([path, "/sounds/winning_sound.wav"]))



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame
from quoridor.client.src.window import pos_in_rect
from quoridor.client.src.colors import Colors


class Player():
    """Create a player"""
    def __init__(self, num_player, walls_remain, orient, color, coord,
                 radius=20):
        self.num_player = num_player
        self.orient = orient
        self.color = color
        self.coord = coord
        self.radius = radius
        self.name = ''
        self.walls_remain = walls_remain

    def set_name(self, name):
        """Set the name of the player"""
        if name != '':
            self.name = name

    def get_num_player(self):
        """Get the num of the player"""
        return self.num_player

    def has_walls(self):
        """Return True if the player has walls"""
        return self.walls_remain > 0

    def can_play(self, game):
        """Return True if the player can play"""
        return game.run and game.current_player == self.num_player

    def can_play_wall(self, game):
        """Return True if the player can play a wall"""
        return self.can_play(game) and self.has_walls()

    def send_move(self, network, coord):
        """Send a move to the server"""
        if self.has_win(coord):
            win = "w"
        else:
            win = "c"
        data = ";".join(['P', str(self.num_player), str(0),
                        str(coord.x), str(coord.y), win])
        network.send(data)

    def play_move(self, walls, network):
        """Play a move if it is possible"""
        keys = pygame.key.get_pressed()
        coord = self.coord

        # Left
        if keys[pygame.K_LEFT]:
            c = coord.west
            if c is not None and walls.no_wall(coord, c):
                if c.is_occuped:
                    c2 = c.west
                    if (c2 is not None and not c2.is_occuped
                            and walls.no_wall(c, c2)):
                        self.send_move(network, c2)
                else:
                    self.send_move(network, c)

        # Right
        elif keys[pygame.K_RIGHT]:
            c = coord.east
            if c is not None and walls.no_wall(coord, c):
                if c.is_occuped:
                    c2 = c.east
                    if (c2 is not None and not c2.is_occuped
                            and walls.no_wall(c, c2)):
                        self.send_move(network, c2)
                else:
                    self.send_move(network, c)

        # Up
        elif keys[pygame.K_UP]:
            c = coord.north
            if c is not None and walls.no_wall(coord, c):
                if c.is_occuped:
                    c2 = c.north
                    if (c2 is not None and not c2.is_occuped
                            and walls.no_wall(c, c2)):
                        self.send_move(network, c2)
                else:
                    self.send_move(network, c)

        # Down
        elif keys[pygame.K_DOWN]:
            c = coord.south
            if c is not None and walls.no_wall(coord, c):
                if c.is_occuped:
                    c2 = c.south
                    if (c2 is not None and not c2.is_occuped
                            and walls.no_wall(c, c2)):
                        self.send_move(network, c2)
                else:
                    self.send_move(network, c)

    def send_wall(self, network, coord, orient):
        """Send a wall to the server"""
        data = ";".join(['P', str(self.num_player), str(1),
                        str(coord.x), str(coord.y), orient])
        network.send(data)

    def play_put_wall(self, pos, coords, walls, network,
                      path_finder, players):
        """Put a wall if it is possible"""
        for c in coords.coords:
            wall_east = c.wall_east
            wall_south = c.wall_south
            for w in [wall_east, wall_south]:
                if (w is not None and pos_in_rect(w.rect_small, pos)
                        and walls.can_add(w)):
                    if path_finder.play_wall(w, players):
                        self.send_wall(network, c, w.orient)
                        return ''
                    else:
                        return "You can't block players!"
        return ''

    def draw(self, win):
        """Draw player on the game board"""
        (x, y) = self.coord.middle
        pygame.draw.circle(win.win, self.color,
                           (x, y), self.radius)
        font = pygame.font.SysFont("comicsans", 40)
        text = font.render(self.name[0], 1, Colors.white)
        win.win.blit(text, (x - self.radius // 2, y - self.radius // 2))

    def has_win(self, coord):
        """Return True if the player wins in the coord"""
        if self.orient == "north" and coord.y == 8:
            return True
        if self.orient == "east" and coord.x == 0:
            return True
        if self.orient == "south" and coord.y == 0:
            return True
        if self.orient == "west" and coord.x == 8:
            return True
        return False


class Players:
    """Manage the players"""
    def __init__(self, nb_players, coords):
        self.nb_players = nb_players
        if nb_players == 4:
            self.players = [
                Player(0, 5, "north", Colors.red, coords.find_coord(4, 0)),
                Player(1, 5, "east", Colors.blue, coords.find_coord(8, 4)),
                Player(2, 5, "south", Colors.green, coords.find_coord(4, 8)),
                Player(3, 5, "west", Colors.yellow, coords.find_coord(0, 4))]
        elif self.nb_players == 2:
            self.players = [
                Player(0, 10, "north", Colors.red, coords.find_coord(4, 0)),
                Player(1, 10, "south", Colors.green, coords.find_coord(4, 8))]
        elif self.nb_players == 3:
            self.players = [
                Player(0, 7, "north", Colors.red, coords.find_coord(4, 0)),
                Player(1, 7, "east", Colors.blue, coords.find_coord(8, 4)),
                Player(2, 7, "south", Colors.green, coords.find_coord(4, 8))]

    def draw(self, win):
        """Draw all players on the game board"""
        for p in self.players:
            if p.name != '':
                p.draw(win)

    def get_player(self, num_player):
        """Get a player"""
        return self.players[num_player]

    def play(self, last_play, coords, walls, path_finder):
        """Make a player play"""
        data = last_play.split(";")
        player = self.get_player(int(data[1]))
        type_play = int(data[2])
        x, y = int(data[3]), int(data[4])
        if type_play == 0:   # Move
            player.coord.is_occuped = False
            player.coord = coords.find_coord(x, y)
            player.coord.is_occuped = True
            win = data[5]
            if win == "w":
                return False
            return True
        elif type_play == 1:    # Wall
            orient = data[5]
            coord_wall = coords.find_coord(x, y)
            if orient == "e":
                wall = coord_wall.wall_east
            elif orient == "s":
                wall = coord_wall.wall_south
            wall.set_color(player.color)
            walls.add_wall(wall)
            player.walls_remain -= 1
            path_finder.add_wall(wall)
        return True

    def set_names(self, names):
        """Set the names of players"""
        for player, name in zip(self.players, names):
            player.set_name(name)

    def reset(self, coords):
        """Reset the players"""
        if self.nb_players == 2:
            walls_remain = 10
        elif self.nb_players == 3:
            walls_remain = 7
        elif self.nb_players == 4:
            walls_remain = 5
        for p in self.players:
            if p.orient == "north":
                p.coord = coords.find_coord(4, 0)
            elif p.orient == "east":
                p.coord = coords.find_coord(8, 4)
            elif p.orient == "south":
                p.coord = coords.find_coord(4, 8)
            elif p.orient == "west":
                p.coord = coords.find_coord(0, 4)
            p.walls_remain = walls_remain



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame


class Wall:
    """Create a wall"""
    def __init__(self, coord1, coord2, win):
        self.coord1 = coord1
        self.coord2 = coord2
        if coord1.same_row(coord2):
            self.coord3 = coord1.south
            self.coord4 = coord2.south
            self.orient = "e"
        if coord1.same_column(coord2):
            self.coord3 = coord1.east
            self.coord4 = coord2.east
            self.orient = "s"

        self.cross_wall = None

        # Window attributs
        self.win = win
        self.rect = self.make_rect()
        self.rect_small = self.make_rect_small()
        self.color = None
        self.code = ";".join([str(coord1.x), str(coord1.y), self.orient])

    def __str__(self):
        return ", ".join([str(self.coord1), str(self.coord2),
                         str(self.coord3), str(self.coord4), self.orient])

    def set_color(self, color):
        """Set the wall's color"""
        self.color = color

    def make_cross_wall(self):
        """Return the cross wall"""
        if self.orient == "e":
            self.cross_wall = self.coord1.wall_south
        if self.orient == "s":
            self.cross_wall = self.coord1.wall_east

    def make_rect(self):
        """Return the rect corresponding to the wall on the window"""
        win = self.win
        (x, y) = self.coord1.top_left
        if self.orient == "e":
            return (x + win.case_side, y,
                    win.wall_width, 2*win.case_side + win.wall_width)
        elif self.orient == "s":
            return (x, y + win.case_side,
                    2*win.case_side + win.wall_width, win.wall_width)
        return None

    def make_rect_small(self):
        """Return the small rect corresponding to the wall on the window"""
        win = self.win
        (x, y) = self.coord1.top_left
        if self.orient == "e":
            return (x + win.case_side, y, win.wall_width, win.case_side)
        elif self.orient == "s":
            return (x, y + win.case_side, win.case_side, win.wall_width)
        return None

    def draw(self, color):
        """Draw the wall on the window"""
        pygame.draw.rect(self.win.win, color, self.rect)

    def __eq__(self, other):
        """Operator == for two walls"""
        return (self.coord1 == other.coord1
                and self.coord2 == other.coord2
                and self.coord3 == other.coord3
                and self.coord4 == other.coord4)


class Walls:
    """Manage all the walls played"""
    def __init__(self):
        self.walls = []
        self.blocked_coords = {}

    def add_wall(self, wall):
        """Add a wall"""
        self.walls.append(wall)
        d = self.blocked_coords
        if wall.coord1.tuple not in d:
            d[wall.coord1.tuple] = [wall.coord2.tuple]
        else:
            d[wall.coord1.tuple].append(wall.coord2.tuple)
        if wall.coord2.tuple not in d:
            d[wall.coord2.tuple] = [wall.coord1.tuple]
        else:
            d[wall.coord2.tuple].append(wall.coord1.tuple)
        if wall.coord3.tuple not in d:
            d[wall.coord3.tuple] = [wall.coord4.tuple]
        else:
            d[wall.coord3.tuple].append(wall.coord4.tuple)
        if wall.coord4.tuple not in d:
            d[wall.coord4.tuple] = [wall.coord3.tuple]
        else:
            d[wall.coord4.tuple].append(wall.coord3.tuple)

    def draw(self):
        """Draw the walls"""
        for w in self.walls:
            w.draw(w.color)

    def contains(self, wall):
        """Return True if wall can't be added"""
        d = self.blocked_coords
        if wall.coord1.tuple in d:
            if wall.coord2.tuple in d[wall.coord1.tuple]:
                return True
        if wall.coord3.tuple in d:
            if wall.coord4.tuple in d[wall.coord3.tuple]:
                return True
        return False

    def wall_in_walls(self, wall):
        """Return True if the wall is in walls"""
        for w in self.walls:
            if wall == w:
                return True
        return False

    def can_add(self, wall):
        """Return True if the wall can be added"""
        return (not self.contains(wall)
                and not self.wall_in_walls(wall.cross_wall))

    def no_wall(self, coord1, coord2):
        """Return True is there is no wall between two positions"""
        d = self.blocked_coords
        if coord1.tuple in d:
            return coord2.tuple not in d[coord1.tuple]
        return True

    def reset(self):
        """Reset the walls"""
        for w in self.walls:
            w.color = None
        self.walls.clear()
        self.blocked_coords.clear()



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame
from quoridor.client.src.wall import Wall


class Coord:
    """Create a coord"""
    def __init__(self, x, y, win, coords):
        self.win = win
        self.coords = coords

        self.x = x
        self.y = y
        self.tuple = (x, y)
        self.is_occuped = False

        # Window attributs
        self.top_left = self.make_top_left()
        self.middle = self.make_middle()
        self.rect = self.make_rect()

        # Links
        self.north = None
        self.east = None
        self.south = None
        self.west = None
        self.wall_east = None
        self.wall_south = None

    def coord_north(self):
        """Return the coord on the north"""
        if self.y - 1 >= 0:
            return self.coords.find_coord(self.x, self.y - 1)
        return None

    def coord_east(self):
        """Return the coord on the east"""
        if self.x + 1 <= 8:
            return self.coords.find_coord(self.x + 1, self.y)
        return None

    def coord_south(self):
        """Return the coord on the south"""
        if self.y + 1 <= 8:
            return self.coords.find_coord(self.x, self.y + 1)
        return None

    def coord_west(self):
        """Return the coord on the west"""
        if self.x - 1 >= 0:
            return self.coords.find_coord(self.x - 1, self.y)
        return None

    def make_top_left(self):
        """Return the top left point of a coord on a window"""
        win = self.win
        x = ((win.wall_width + win.case_side)*self.x
             + win.wall_width + win.top_left[0])
        y = ((win.wall_width + win.case_side)*self.y
             + win.wall_width + win.top_left[1])
        return (x, y)

    def make_middle(self):
        """Return the middle point of a coord on a window"""
        win = self.win
        x = ((win.wall_width + win.case_side)*self.x
             + (win.wall_width + win.case_side // 2)
             + win.top_left[0])
        y = ((win.wall_width + win.case_side)*self.y
             + (win.wall_width + win.case_side // 2)
             + win.top_left[1])
        return (x, y)

    def make_rect(self):
        """Return the rectangle of the coord"""
        win = self.win
        x, y = self.top_left
        return (x, y, win.case_side, win.case_side)

    def make_wall_east(self):
        """Return the east wall of the coord"""
        if self.east is not None and self.y != 8:
            return Wall(self, self.east, self.win)
        return None

    def make_wall_south(self):
        """Return the south wall of the coord"""
        if self.south is not None and self.x != 8:
            return Wall(self, self.south, self.win)
        return None

    def link_coord(self):
        """Link the coords"""
        self.north = self.coord_north()
        self.east = self.coord_east()
        self.south = self.coord_south()
        self.west = self.coord_west()

    def make_walls(self):
        """Make the walls around the coord"""
        self.wall_east = self.make_wall_east()
        self.wall_south = self.make_wall_south()

    def make_cross_walls(self):
        """Make the cross walls of the walls of the coord"""
        if self.wall_east is not None:
            self.wall_east.make_cross_wall()
        if self.wall_south is not None:
            self.wall_south.make_cross_wall()

    def same_row(self, other):
        """Return True if the two coords are on the same row"""
        return self.y == other.y

    def same_column(self, other):
        """Return True if the two coords are on the same column"""
        return self.x == other.x

    def __str__(self):
        """String format of a coord"""
        return f"({self.x}, {self.y})"

    def __eq__(self, other):
        """Operator == between two coords"""
        return self.x == other.x and self.y == other.y

    def draw(self, color):
        """Draw the rectangle of a coord"""
        pygame.draw.rect(self.win.win, color, self.rect)


class Coords:
    """Manage the coords"""
    def __init__(self, win):
        self.win = win
        self.coords = self.make_coords()
        self.link_coords()
        self.make_walls()

    def make_coords(self):
        """Make coords"""
        coords = []
        for x in range(9):
            for y in range(9):
                coords.append(Coord(x, y, self.win, self))
        return coords

    def link_coords(self):
        """Link coords"""
        for c in self.coords:
            c.link_coord()

    def make_walls(self):
        """Make walls"""
        for c in self.coords:
            c.make_walls()
        for c in self.coords:
            c.make_cross_walls()

    def find_coord(self, x, y):
        """Find the coord corresponding to x and y"""
        return self.coords[x * 9 + y]

    def reset(self):
        """Reset coords"""
        for c in self.coords:
            c.is_occuped = False



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import sys
import os
import pygame
from quoridor.client.src.network import Network
from quoridor.client.src.window import Window
from quoridor.client.src.player import Players
from quoridor.client.src.wall import Walls
from quoridor.client.src.pathfinder import PathFinder
from quoridor.client.src.sounds import Sounds


def client(host, port):
    """Game client"""
    print("Welcome to Quoridor Online!")
    print("Host:", host)
    print("Port:", port)
    # Name input
    name = ''
    while name == '' or len(name) > 10:
        name = input("Enter your name: ").capitalize()
        if len(name) > 10:
            print("Too long")
    print("Waiting for the server...")

    # Connexion
    n = Network(host, port)
    try:
        n.connect()
        # Receive nb players, num player
        data = n.recv()
        nb_players, num_player = int(data[0]), int(data[1])
    except Exception:
        print("Unable to connect to the server")
        exit()

    # Send name
    try:
        n.send(';'.join(['N', str(num_player), name]))
    except Exception:
        print("Impossible to send data to the server")
        exit()

    print("Connexion established!")
    print(f"Hello {name}! You are player {num_player}!")

    # Init pygame
    pygame.init()
    clock = pygame.time.Clock()
    path = os.path.dirname(os.path.abspath(__file__))
    sounds = Sounds(path)

    # Init game
    win = Window()
    coords = win.coords
    players = Players(nb_players, coords)
    player = players.players[num_player]
    walls = Walls()
    pf = PathFinder()

    last_play = ''
    run = True
    ready = False
    start = False

    while run:
        clock.tick(40)
        n.send('get')
        try:
            game = n.get_game()
        except Exception:
            run = False
            print("Couldn't get game")
            break

        # Start a game
        if not start:
            if not ready:
                players.set_names(game.names)
                win.update_info(
                    f"Waiting for {game.players_missing()} players...")
                ready = game.ready()
            if game.run:
                current_p = players.players[game.current_player]
                win.update_info(f"Let's go! {current_p.name} plays!",
                                current_p.color)
                try:
                    sounds.start_sound.play()
                except Exception:
                    pass
                start = True
            elif game.wanted_restart != []:
                nb = len(game.wanted_restart)
                p = players.players[game.wanted_restart[-1]]
                win.update_info(
                    f"{p.name} wants to restart! ({nb}/{nb_players})",
                    p.color)

        # Continue a game
        else:
            get_play = game.last_play
            if get_play != '' and get_play != last_play:
                if get_play[0] == 'D':
                    current_p = players.players[game.current_player]
                    win.update_info(f"{current_p.name} plays!",
                                    current_p.color)
                elif get_play[0] == 'P':
                    if players.play(get_play, coords, walls, pf):
                        current_p = players.players[game.current_player]
                        win.update_info(f"{current_p.name} plays!",
                                        current_p.color)
                    else:
                        win.update_info(f"{game.winner} wins!")
                        try:
                            sounds.winning_sound.play()
                        except Exception:
                            pass
                        win.button_restart.show = True
                        start = False
                last_play = get_play

        pos = pygame.mouse.get_pos()
        win.redraw_window(game, players, walls, pos)

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                run = False
                pygame.quit()

            elif event.type == pygame.MOUSEBUTTONDOWN:
                pos = pygame.mouse.get_pos()
                if win.button_quit.click(pos):
                    run = False
                    pygame.quit()
                elif (not game.run and win.button_restart.click(pos)
                        and win.button_restart.show):   # Restart
                    coords.reset()
                    players.reset(coords)
                    walls.reset()
                    pf.reset()
                    win.button_restart.show = False
                    n.send(';'.join(['R', str(num_player)]))
                elif player.can_play_wall(game):    # Put a wall
                    mes = player.play_put_wall(
                        pos, coords, walls, n, pf, players)
                    if mes != '':
                        win.update_info(mes)

            elif event.type == pygame.KEYDOWN:  # Move pawn
                if player.can_play(game):
                    player.play_move(walls, n)

    print("Good bye!")


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Not enough arguments: [Host address] [Port number]")
    else:
        host = str(sys.argv[1])
        port = int(sys.argv[2])
        client(host, port)



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import socket
import pickle


class Network:
    """Manage communication between client and server"""
    def __init__(self, host, port):
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host = host
        self.port = port
        self.addr = (self.host, self.port)

    def connect(self):
        """Connect to the server"""
        try:
            self.client.connect(self.addr)
        except socket.error as e:
            print(e)

    def recv(self):
        """Receive data from server"""
        try:
            return self.client.recv(2048).decode().split(';')
        except socket.error as e:
            print(e)

    def send(self, data):
        """Send data to server"""
        try:
            self.client.send(str.encode(data))
        except socket.error as e:
            print(e)

    def get_game(self):
        """Get the game"""
        return pickle.loads(self.client.recv(2048))



"""
Quoridor Online
Quentin Deschamps, 2020
"""
from pathfinding.core.grid import Grid
from pathfinding.finder.a_star import AStarFinder


class PathFinder:
    """Create the path finder to avoid blocked players"""
    def __init__(self):
        self.finder = AStarFinder()
        self.side = 17
        self.matrix = [[1]*self.side for _ in range(self.side)]
        self.make_walls()

    def make_walls(self):
        """Init the walls"""
        for i in range(self.side):
            if i % 2 == 1:
                for j in range(self.side):
                    if j % 2 == 1:
                        self.matrix[i][j] = 0

    def pos_player_in_grid(self, player):
        """Return the position of a player in the grid"""
        return (2*player.coord.y, 2*player.coord.x)

    def print_grid(self, grid=True, players=None):
        """Print the grid"""
        matrix = self.matrix.copy()
        if players is not None:
            for p in players.players:
                i, j = self.pos_player_in_grid(p)
                matrix[i][j] = -1
        grid = Grid(matrix=matrix)
        print(grid.grid_str())

    def add_wall(self, wall):
        """Add a wall"""
        if wall.orient == "e":
            y, x = 2*wall.coord1.x + 1, 2*wall.coord1.y
            for i in range(3):
                self.matrix[x + i][y] = 0
        elif wall.orient == "s":
            y, x = 2*wall.coord1.x, 2*wall.coord1.y + 1
            for i in range(3):
                self.matrix[x][y + i] = 0

    def remove_wall(self, wall):
        """Remove a wall"""
        if wall.orient == "e":
            y, x = 2*wall.coord1.x + 1, 2*wall.coord1.y
            for i in range(3):
                self.matrix[x + i][y] = 1
        elif wall.orient == "s":
            y, x = 2*wall.coord1.x, 2*wall.coord1.y + 1
            for i in range(3):
                self.matrix[x][y + i] = 1

    def find_path(self, player, show=False):
        """Return True if the player can finish"""
        x, y = self.pos_player_in_grid(player)
        grid = Grid(matrix=self.matrix)
        if player.orient == "north":
            x_end = self.side - 1
            for y_end in range(0, self.side, 2):
                start = grid.node(y, x)
                end = grid.node(y_end, x_end)
                path, runs = self.finder.find_path(start, end, grid)
                if path != []:
                    if show:
                        print(grid.grid_str(path=path, start=start, end=end))
                    return True
                grid.cleanup()

        elif player.orient == "east":
            y_end = 0
            for x_end in range(0, self.side, 2):
                start = grid.node(y, x)
                end = grid.node(y_end, x_end)
                path, runs = self.finder.find_path(start, end, grid)
                if path != []:
                    if show:
                        print(grid.grid_str(path=path, start=start, end=end))
                    return True
                grid.cleanup()

        elif player.orient == "south":
            x_end = 0
            for y_end in range(0, self.side, 2):
                start = grid.node(y, x)
                end = grid.node(y_end, x_end)
                path, runs = self.finder.find_path(start, end, grid)
                if path != []:
                    if show:
                        print(grid.grid_str(path=path, start=start, end=end))
                    return True
                grid.cleanup()

        elif player.orient == "west":
            y_end = self.side - 1
            for x_end in range(0, self.side, 2):
                start = grid.node(y, x)
                end = grid.node(y_end, x_end)
                path, runs = self.finder.find_path(start, end, grid)
                if path != []:
                    if show:
                        print(grid.grid_str(path=path, start=start, end=end))
                    return True
                grid.cleanup()

        return False

    def play_wall(self, wall, players):
        """Return True if the wall doesn't block players"""
        self.add_wall(wall)
        for p in players.players:
            if not self.find_path(p):
                self.remove_wall(wall)
                return False
        return True

    def reset(self):
        """Reset the matrix"""
        self.matrix.clear()
        self.matrix = [[1]*self.side for _ in range(self.side)]
        self.make_walls()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quoridor",
    version="0.0.2",
    author="Quentin Deschamps",
    author_email="quentindeschamps18@gmail.com",
    description="Quoridor Online Game",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Quentin18/Quoridor-Online",
    packages=["quoridor", "quoridor.client", "quoridor.client.src",
              "quoridor.server", "quoridor.server.src"],
    include_package_data=True,
    install_requires=['pygame', 'pathfinding'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Quoridor Online
Quentin Deschamps, 2020
"""
import pygame
from quoridor.client.src.colors import Colors


class Text:
    """Create a text box"""
    def __init__(self, text, color, size=30, font="comicsans"):
        self.text = text
        self.color = color
        self.size = size
        self.font = font

    def draw(self, win, pos):
        """Draw the text on the window"""
        font = pygame.font.SysFont(self.font, self.size)
        text = font.render(self.text, 1, self.color, True)
        win.blit(text, pos)


class Button:
    """Create a button"""
    def __init__(self, text, x, y, color, width=150, height=40, show=True):
        self.text = text
        self.x = x
        self.y = y
        self.color = color
        self.width = width
        self.height = height
        self.show = show

    def draw(self, win):
        """Draw the button on the window"""
        pygame.draw.rect(win, self.color,
                         (self.x, self.y, self.width, self.height))
        font = pygame.font.SysFont("comicsans", 40)
        text = font.render(self.text, 1, Colors.black)
        win.blit(text,
                 (self.x + round(self.width/2) - round(text.get_width()/2),
                  self.y + round(self.height/2) - round(text.get_height()/2)))

    def click(self, pos):
        """Return True if pos is in the button rectangle"""
        pos_x, pos_y = pos
        return (self.x <= pos_x <= self.x + self.width
                and self.y <= pos_y <= self.y + self.height)
